from pathlib import Path
from cryptography.fernet import Fernet
from os import environ

_token = b'gAAAAABifAIMNFaSNF8epJIDWIv2nSe3zxARkMmViCa1ZCvtwoRqhuB1LYjjJsAstwTvP4dEOSm6Wj0SRDWr3PPwZz5eEBt_1fU8uIaninakGYFNSarEduD6YfoA-rm28qUQHYpVcuae3lj8sYrs_87P6F4s3gBrYg=='
_key = b'ywE5qRot_nuWfLnbEXXcAPKaW10us3YpWEkDXgm9was='

CURSEFORGE_API_TOKEN = Fernet(_key).decrypt(_token).decode()
OAUTH_GITHUB_CLIENT_ID = "8011f22f502b091464de"

CACHE_HOME = environ.get("XDG_CACHE_HOME", Path().home() / ".cache")
DEFAULT_CACHE_DIR = Path(CACHE_HOME) / "mmc-export"

output_naming_scheme = "{abbr}_{name}"


from configparser import ConfigParser
from json import loads as parse_json
from pathlib import Path

from aiohttp_client_cache.session import CachedSession

from .Helpers.resourceAPI import ResourceAPI_Batched
from .Helpers.structures import File, Format, Intermediate
from .Helpers.utils import get_hash


class Parser(Format):

    def __init__(self, path: Path, session: CachedSession) -> None:

        self.intermediate = Intermediate()
        self.resourceAPI = ResourceAPI_Batched(session, self.intermediate)

        super().__init__(path)

    def get_basic_info(self):

        data = next(self.temp_dir.glob("**/instance.cfg")).read_text()

        cfg = ConfigParser()
        cfg.read_string("[dummy_section]\n" + data)
        if name := cfg['dummy_section'].get('name'):
            self.intermediate.name = name

        bdata = next(self.temp_dir.glob("**/mmc-pack.json")).read_bytes()
        pack_info = parse_json(bdata)        

        for component in pack_info['components']:

            match component:

                case {'cachedName': "Minecraft", 'version': version}: 
                    self.intermediate.minecraft_version = version
                case {'cachedName': "Fabric Loader", 'version': version}: 
                    self.intermediate.modloader.type = "fabric"
                    self.intermediate.modloader.version = version
                case {'cachedName': "Quilt Loader", 'version': version}: 
                    self.intermediate.modloader.type = "quilt"
                    self.intermediate.modloader.version = version
                case {'cachedName': "Forge", 'version': version}: 
                    self.intermediate.modloader.type = "forge"
                    self.intermediate.modloader.version = version

    def get_override(self, path: Path):

        if "minecraft" not in path.parts: return
        root_dir_id = path.parts.index("minecraft")
        relative_path = path.relative_to(*path.parts[:root_dir_id + 1]).parent

        file = File(
            name = path.name,
            hash = File.Hash(sha256=get_hash(path)),
            path = path,
            relativePath = relative_path.as_posix())
        
        self.intermediate.overrides.append(file)

    async def parse(self) -> Intermediate:
        
        downloadable_content = ("resourcepacks", "shaderpacks", "mods")

        from shutil import unpack_archive        
        unpack_archive(self.modpack_path, self.temp_dir)
        self.get_basic_info()

        overrides = list()

        for file in [file for file in self.temp_dir.glob("**/*") if file.is_file()]:
            if file.parent.name in downloadable_content and file.suffix != ".txt": 
                self.resourceAPI.queue_resource(file)
            else: overrides.append(file)

        self.intermediate.resources = await self.resourceAPI.gather()

        for override in overrides:
            self.get_override(override)

        return self.intermediate



from pathlib import Path

from pytoml import dump as write_toml

from ..Helpers.structures import File, Intermediate, Resource, Writer
from ..Helpers.utils import get_hash, get_name_from_scheme


class packwiz(Writer):

    def __init__(self, path: Path, intermediate: Intermediate) -> None:

        self.pack_info = dict()

        self.index = {
            "hash-format": "sha256",
            "files": []
        }

        super().__init__(path, intermediate)

    def add_resource(self, resource: Resource) -> None:

        if not resource.providers: return self.add_override(resource.file)
    
        data = {
            "name": resource.name,
            "filename": resource.file.name,

            "update": {}
        }

        slug = None

        if provider := resource.providers.get("CurseForge"):

            slug = provider.slug
            
            data['update']['curseforge'] = {
                "file-id": provider.fileID,
                "project-id": provider.ID,
                "release-channel": "beta"
            }

            data['download'] = {
                "hash-format": "sha1",
                "hash": resource.file.hash.sha1,
                "mode": "metadata:curseforge"
            }

        elif provider := resource.providers.get("Modrinth"):

            slug = provider.slug

            data['update']['modrinth'] = {
                "mod-id": provider.ID,
                "version": provider.fileID
            }

            data['download'] = {
                "url": provider.url,
                "hash-format": "sha512",
                "hash": resource.file.hash.sha512
            }

        elif provider := resource.providers.get("Other"):

            slug = provider.slug

            data['download'] = {
                "url": provider.url,
                "hash-format": "sha256",
                "hash": resource.file.hash.sha256
            }

        if resource.optional: data['option'] = {"optional": True}

        from werkzeug.utils import secure_filename
        if not slug: slug = secure_filename(resource.name)

        toml_path = self.temp_dir / resource.file.relativePath / (slug + ".toml")
        toml_path.parent.mkdir(parents=True, exist_ok=True)

        with open(toml_path, "w") as file:
            write_toml(data, file)

        index_data = {
            "file": toml_path.relative_to(self.temp_dir).as_posix(),
            "hash": get_hash(toml_path),
            "metafile": True
        }
        
        self.index['files'].append(index_data)

    def add_override(self, file: File) -> None:

        file_path = self.temp_dir / file.relativePath
        file_path.mkdir(parents=True, exist_ok=True)

        from shutil import copy2 as copy_file
        copy_file(file.path, file_path)

        data = {
            "file": Path(file.relativePath) / file.name,
            "hash": file.hash.sha256
        }
        
        self.index['files'].append(data)

    def write(self) -> None:

        for override in self.intermediate.overrides:
            self.add_override(override)

        for resource in self.intermediate.resources:
            self.add_resource(resource)

        index_path = self.temp_dir / "index.toml"
        with open(index_path, "w") as file:
            write_toml(self.index, file)

        self.pack_info = {
            "name": self.intermediate.name,
            "author": self.intermediate.author,
            "version": self.intermediate.version,

            "index": {
                "file": "index.toml",
                "hash-format": "sha256",
                "hash": get_hash(index_path)
            },

            "versions": {
                self.intermediate.modloader.type: self.intermediate.modloader.version,
                "minecraft": self.intermediate.minecraft_version
            }
        }

        with open(self.temp_dir / "pack.toml", "w") as file:
            write_toml(self.pack_info, file)

        from shutil import make_archive
        name = get_name_from_scheme("PW", "Packwiz", self.intermediate)
        make_archive((self.modpack_path / name).as_posix(), 'zip', self.temp_dir, '.')



from importlib import import_module
from json import dump as write_json
from shutil import rmtree

from aiohttp import TCPConnector
from aiohttp_client_cache.backends.filesystem import FileBackend
from aiohttp_client_cache.session import CachedSession

from .Helpers.resourceAPI import ResourceAPI
from .Helpers.utils import (JsonEncoder, add_github_token, parse_args,
                            read_config_into, resolve_conflicts)
from .parser import Parser
from . import config


async def program():

    args = parse_args()

    ResourceAPI.modrinth_search_type = args.modrinth_search
    ResourceAPI.excluded_providers = args.excluded_providers

    cache = FileBackend("mmc-export", use_temp=True, urls_expire_after={'*.jar': -1}, allowed_methods=("GET", "POST", "HEAD"))
    async with CachedSession(cache=cache, connector=TCPConnector(limit=0)) as session: 
        if args.skip_cache: session.cache.disabled = True # type: ignore

        match args.cmd:
            case "gh-login": await add_github_token(session); return # type: ignore
            case "gh-logout": 
                url = "https://github.com/settings/connections/applications/8011f22f502b091464de"
                print(f"You can revoke your access token by the following link: \n{url}"); return
            case "purge-cache":
                if args.cache_web or args.cache_all: await session.cache.clear() # type: ignore
                if args.cache_files or args.cache_all: rmtree(config.DEFAULT_CACHE_DIR, ignore_errors=True)
                return

        parser = Parser(args.input, session) # type: ignore
        intermediate = await parser.parse()
        
        if version := args.modpack_version: intermediate.version = version
        read_config_into(args.config, intermediate)
        await resolve_conflicts(session, intermediate) # type: ignore

        for format in args.formats:

            if format == "Intermediate":
                with open(args.output / "intermediate_output.json", "w") as file:
                    write_json(intermediate, file, indent=4, cls=JsonEncoder)
                continue

            module = import_module(f".Formats.{format.lower()}", "mmc_export")
            Writer = getattr(module, format)

            writer = Writer(args.output, intermediate)
            writer.write()         

    return 0

def main():
    import asyncio, sys
    if sys.platform.startswith("win"):
        policy = asyncio.WindowsSelectorEventLoopPolicy()  # type: ignore
        asyncio.set_event_loop_policy(policy)
                
    try: asyncio.run(program())
    except KeyboardInterrupt: 
        print("Operation aborted by user.")



import asyncio, sys, re
import tenacity as tn
from io import BytesIO
from typing import Any
from pathlib import Path
from pprint import pformat
from urllib.parse import urlparse

import keyring as secret_store
from argparse import SUPPRESS, ArgumentParser, Namespace
from aiohttp_client_cache.session import CachedSession
from pytoml import loads as parse_toml

from .. import config
from .structures import Intermediate, Resource


def get_hash(file: Path | BytesIO | bytes, hash_type: str = "sha256") -> str:

    if isinstance(file, Path): data = file.read_bytes()
    elif isinstance(file, BytesIO): data = file.read(); file.seek(0)
    elif isinstance(file, bytes): data = file
    else: raise TypeError("Incorrect file type!")
        
    from hashlib import sha1, sha256, sha512
    from murmurhash2 import murmurhash2 as murmur2
    from xxhash import xxh3_64_hexdigest

    match hash_type:
        case "sha1": hash = sha1(data).hexdigest()
        case "sha256": hash = sha256(data).hexdigest()
        case "sha512": hash = sha512(data).hexdigest()
        case "xxhash": hash = xxh3_64_hexdigest(data)
        case "murmur2": 
            data = re.sub(rb"[\x09\x0A\x0D\x20]", b"", data)
            hash = str(murmur2(data, seed=1))
        case _: raise TypeError("Incorrect hash type!")

    return hash

def get_hashes(file: Path | BytesIO | bytes, *args: str):
    return [get_hash(file, hash_type) for hash_type in args]

async def add_github_token(session: CachedSession) -> None:

    headers = {"Accept": "application/json"}

    async with session.disabled():
        url = "https://github.com/login/device/code"
        async with session.post(url, params={"client_id": config.OAUTH_GITHUB_CLIENT_ID}, headers=headers) as response:
            data = await response.json()

            device_code = data['device_code']
            user_code = data['user_code']
            verification_uri = data['verification_uri']
            interval = data['interval']

            print(f" To proceed, enter the code: {user_code}")
            print(f"Here: {verification_uri}")

        payload = {"client_id": config.OAUTH_GITHUB_CLIENT_ID, "device_code": device_code, 
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code"}

        while(True):      
            url = "https://github.com/login/oauth/access_token"
            async with session.post(url, params=payload, headers=headers) as response:
                data = await response.json()

                match data.get('error'):
                    case "authorization_pending": await asyncio.sleep(interval); continue
                    case "expired_token": print("Confirmation time is expired"); return
                    case "access_denied": print("Token request declined"); return

                if token := data.get('access_token'):
                    secret_store.set_password("mmc-export", "github-token", token)
                    print("Successfully authorized!")
                    return

def delete_github_token() -> None:
    try: secret_store.delete_password("mmc-export", "github-token")
    except secret_store.core.backend.errors.PasswordDeleteError: return

def get_github_token() -> str | None:
    return secret_store.get_password("mmc-export", "github-token")

def parse_args() -> Namespace:

    formats = ('packwiz', 'Modrinth', 'CurseForge', 'Intermediate')
    mr_search = ('exact', 'accurate', 'loose')
    providers = ('GitHub', 'CurseForge', 'Modrinth')
 
    arg_parser = ArgumentParser(usage=SUPPRESS, add_help=False)
    arg_parser.add_argument('-h', '--help', dest="help", action='store_true')
    arg_parser.add_argument('-c', '--config', dest='config', type=Path)
    arg_parser.add_argument('-i', '--input', dest='input', type=Path)
    arg_parser.add_argument('-f', '--format', dest='formats', type=str, nargs="+", choices=formats)
    arg_parser.add_argument('-o', '--output', dest='output', type=Path, default=Path.cwd())
    arg_parser.add_argument('--modrinth-search', dest='modrinth_search', type=str, choices=mr_search, default='exact')
    arg_parser.add_argument('--exclude-providers', dest='excluded_providers', type=str, nargs="+", choices=providers, default=[])
    arg_parser.add_argument('--skip-cache', dest='skip_cache', action='store_true')
    arg_parser.add_argument('-v', '--version', dest='modpack_version', type=str)
    arg_parser.add_argument('--scheme', dest='scheme', type=str)

    arg_subs = arg_parser.add_subparsers(dest='cmd')
    arg_subs.add_parser('gh-login', add_help=False)
    arg_subs.add_parser('gh-logout', add_help=False)

    arg_cache = arg_subs.add_parser('purge-cache', add_help=False)
    arg_cache.add_argument('--web', dest='cache_web', action='store_true')
    arg_cache.add_argument('--files', dest='cache_files', action='store_true')
    arg_cache.add_argument('--all', dest='cache_all', action='store_true')
    
    args = arg_parser.parse_args(args=None if sys.argv[1:] else ['--help'])

    if args.help: 
        print("mmc-export: Export MMC modpack to other modpack formats")
        print("Usage example you can find here: https://github.com/RozeFound/mmc-export#how-to-use")

    if args.scheme: 
        old_scheme = config.output_naming_scheme
        config.output_naming_scheme = args.scheme
        try: get_name_from_scheme("PRAB", "FormatName", Intermediate("ModpackName", version="0.0.0"))
        except KeyError as e:
            print("Found an error(s) in naming scheme!")
            for arg in e.args: print(f"Keyword {arg} doesn't exists")
            print("Please, recheck your scheme.")
            print("Default will be used for this run.")
            config.output_naming_scheme = old_scheme

    if args.cmd and args.cmd == "purge-cache":
        if not args.cache_web \
            or not args.cache_files \
            or not args.cache_all:
            args.cache_all = True

    if not args.cmd:
        if not args.input: arg_parser.error("Input must be specified!")
        if not args.formats: arg_parser.error("At least one format must be specified!")
        if not args.input.exists(): arg_parser.error("Invalid input!")

    return args

def read_config_into(cfg_path: Path, intermediate: Intermediate) -> None:

    allowed_domains = ("cdn.modrinth.com", "edge.forgecdn.net", "media.forgecdn.net", "gitlab.com", "github.com", "raw.githubusercontent.com")
    lost_resources = [res for res in intermediate.resources if not res.providers]

    if cfg_path is not None and cfg_path.exists():
        config = parse_toml(cfg_path.read_text())
    else: config = {}

    intermediate.name = config.get('name', intermediate.name)
    intermediate.author = config.get('author', intermediate.author)
    intermediate.version = config.get('version', intermediate.version)
    intermediate.description = config.get('description', intermediate.description)
    if not intermediate.version: intermediate.version = input("Specify modpack version: ")

    for resource_config in config.get('Resource', []):

        url = resource_config.get('url')
        name = resource_config.get('name', "")
        filename = resource_config.get('filename', "")

        resource = next((x for x in intermediate.resources if name == x.name or filename == x.file.name), None)
        if not resource: continue

        if resource_config.get("optional") is True:
            resource.optional = True

        match resource_config.get("action"):

            case None: # if action not specified, try to add a provider
                if not url: 
                    print(f"Failed to read config for {resource.name}, you must specify url!")
                elif urlparse(url).netloc not in allowed_domains:
                    print(f"Failed to read config for {resource.name}, wrong url domain!")
                    print(f"Allowed domains: {pformat(allowed_domains)}")
                else: resource.providers['Other'] = Resource.Provider(url = url)

            case "remove": 
                intermediate.resources.remove(resource)
            case "override": 
                intermediate.overrides.append(resource.file)
                intermediate.resources.remove(resource)
            case "ignore": pass
            case _: print(f"Wrong action for {resource.name}!")

        if resource in lost_resources:
            lost_resources.remove(resource)

    for resource in lost_resources:
        print("No config entry found for resource:", resource.name)
        
async def resolve_conflicts(session: CachedSession, intermediate: Intermediate) -> None: 

    @tn.retry(stop=tn.stop.stop_after_attempt(5), wait=tn.wait.wait_fixed(1))
    async def download_file(url: str) -> tuple[str, bytes]:
        async with session.get(url) as response:
            return url, await response.read()

    futures = [download_file(r.providers['Other'].url) for r 
        in intermediate.resources if "Other" in r.providers]
    files = await asyncio.gather(*futures)

    for resource in intermediate.resources:
        if provider := resource.providers.get('Other'):
            cloud_file = next(file for url, file in files if url == provider.url)
            sha1, sha256, sha512 = get_hashes(cloud_file, "sha1", "sha256", "sha512")
            if "Modrinth" in resource.providers:
                if resource.file.hash.sha1 != sha1 or resource.file.hash.sha512 != sha512:
                    resource.providers.pop("Other")
            else: 
                resource.file.hash.sha1 = sha1
                resource.file.hash.sha256 = sha256 
                resource.file.hash.sha512 = sha512 
                resource.file.size = len(cloud_file)

def get_name_from_scheme(abbr: str, format: str, pack: Intermediate) -> str:
    return config.output_naming_scheme.format(abbr=abbr, format=format, name=pack.name, version=pack.version, pack=pack)


import dataclasses, json
class JsonEncoder(json.JSONEncoder):

    def clean(self, value) -> dict | list | Any:
        if isinstance(value, list): return [self.clean(x) for x in value if x]
        elif isinstance(value, dict): return {key: self.clean(val) for key, val in value.items() if val}
        else: return value

    def default(self, o: Any) -> Any:
        
        if dataclasses.is_dataclass(o):
            data = dataclasses.asdict(o)
            return self.clean(data)
        if isinstance(o, Path):
            return o.as_posix()

        return super().default(o)



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from pathlib import Path
from typing import Literal


@dataclass
class File:

    name: str = field(default_factory=str)

    @dataclass
    class Hash:
        sha1: str = field(default_factory=str)
        sha256: str = field(default_factory=str)
        sha512: str = field(default_factory=str)
        murmur2: str = field(default_factory=str)

    hash: Hash = field(default_factory=Hash)
    size: int = field(default_factory=int)

    path: Path = field(default_factory=Path)
    relativePath: str = field(default_factory=str)


@dataclass
class Resource:

    "Represents downloadable item i.e. mod, resourcepack, shaderpack etc."

    name: str = field(default_factory=str)
    links: list[str] = field(default_factory=list)
    optional: bool = False

    @dataclass
    class Provider:

        ID: str | int | None = None
        fileID: str | int | None = None
        url: str = field(default_factory=str)

        slug: str = field(default_factory=str)
        author: str = field(default_factory=str)

    file: File = field(default_factory=File)
    providers: dict[Literal["Modrinth", "CurseForge", "Other"], Provider] = field(default_factory=dict)


@dataclass
class Intermediate:

    name: str = field(default_factory=str)
    author: str = field(default_factory=str)
    version: str = field(default_factory=str)
    description: str = field(default_factory=str)

    @dataclass
    class ModLoader:
        type: str = field(default_factory=str)
        version: str = field(default_factory=str)

    minecraft_version: str = field(default_factory=str)
    modloader: ModLoader = field(default_factory=ModLoader)

    resources: list[Resource] = field(default_factory=list)
    overrides: list[File] = field(default_factory=list)


class Format(ABC):
    def __init__(self, path: Path) -> None:

        from tempfile import TemporaryDirectory

        self._temp_dir = TemporaryDirectory()
        self.temp_dir = Path(self._temp_dir.name)

        self.modpack_path = path

    def __del__(self) -> None:
        self._temp_dir.cleanup()


class Writer(Format):
    def __init__(self, path: Path, intermediate: Intermediate) -> None:

        self.intermediate = intermediate

        super().__init__(path)

    @abstractmethod
    def write(self) -> None:
        raise NotImplementedError()



from pathlib import Path

from ..Helpers.structures import File, Intermediate, Resource, Writer
from ..Helpers.utils import get_name_from_scheme


class CurseForge(Writer):

    def __init__(self, path: Path, intermediate: Intermediate) -> None:

        self.manifest = dict()
        self.modlist = list()

        super().__init__(path, intermediate)

    def add_resource(self, resource: Resource) -> None:

        if provider := resource.providers.get('CurseForge'):

            data = {
                "projectID": provider.ID,
                "fileID": provider.fileID,
                "required": not resource.optional
            }

            self.manifest['files'].append(data)
            mod_page_url = sorted(set(resource.links))[-1]
            self.modlist.append(f"<li><a href=\"{mod_page_url}\">{resource.name} (by {provider.author})</a></li>\n")

        else: self.add_override(resource.file)

    def add_override(self, file: File) -> None:

        overrides_dir = self.temp_dir / "overrides"
        overrides_dir.mkdir(parents=True, exist_ok=True)

        file_path = overrides_dir / file.relativePath
        file_path.mkdir(parents=True, exist_ok=True)

        from shutil import copy2 as copy_file
        copy_file(file.path, file_path)

    def write_manifest(self) -> None:

        self.manifest = {
            
            "minecraft": {
                "version": self.intermediate.minecraft_version,
                "modLoaders": [
                    {
                    "id": f"{self.intermediate.modloader.type}-{self.intermediate.modloader.version}",
                    "primary": True
                    }
                ]
            },
            "manifestType": "minecraftModpack",
            "manifestVersion": 1,
            "name": self.intermediate.name,
            "version": self.intermediate.version,
            "author": self.intermediate.author,

            "files": [],
            "overrides": "overrides"
        }


    def write(self) -> None:

        self.write_manifest()

        for resource in self.intermediate.resources:
            self.add_resource(resource)

        for override in self.intermediate.overrides:
            self.add_override(override)

        from json import dump as write_json
        with open(self.temp_dir / "manifest.json", 'w') as file:
            write_json(self.manifest, file, indent=4)

        with open(self.temp_dir / "modlist.html", 'w') as file:
            self.modlist.insert(0, "<ul>\n")
            self.modlist.append("</ul>\n")
            file.writelines(self.modlist)

        from shutil import make_archive
        name = get_name_from_scheme("CF", "CurseForge", self.intermediate)
        make_archive((self.modpack_path / name).as_posix(), 'zip', self.temp_dir, '.')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mmc_export', 'mmc_export.Formats', 'mmc_export.Helpers']

package_data = \
{'': ['*']}

install_requires = \
['Werkzeug>=2.1.2,<3.0.0',
 'aiohttp-client-cache[filesystem]>=0.7.0,<0.8.0',
 'aiohttp>=3.8.1,<4.0.0',
 'cryptography>=37.0.2,<38.0.0',
 'gql-query-builder>=0.1.7,<0.2.0',
 'keyring>=23.5.1,<24.0.0',
 'murmurhash2>=0.2.10,<0.3.0',
 'pytoml>=0.1.21,<0.2.0',
 'tenacity>=8.0.1,<9.0.0',
 'xxhash>=3.0.0,<4.0.0']

entry_points = \
{'console_scripts': ['mmc-export = mmc_export:main']}

setup_kwargs = {
    'name': 'mmc-export',
    'version': '2.5.9',
    'description': 'Export MMC modpack to other modpack formats',
    'long_description': '# MultiMC advanced exporter\n[![PyPI pyversions](https://img.shields.io/pypi/pyversions/mmc-export)](https://www.python.org)\n[![PyPI version](https://img.shields.io/pypi/v/mmc-export?label=mmc-export&color=%2347a637)](https://pypi.org/project/mmc-export)\n[![PyPI downloads](https://img.shields.io/pypi/dm/mmc-export?color=%23894bbf)](https://pypistats.org/packages/mmc-export)\n[![GitHub license](https://img.shields.io/github/license/RozeFound/mmc-export)](/LICENSE)\n\nSince MultiMC export features are very limited, I created a script that solves this problem, with this script you can export MultiMC pack to any popular format (e.g. curseforge, modrinth, packwiz). MultiMC forks which didn\'t changed export format much also supported, PolyMC support approved.\n\n# Features\n\n- Support conversion to:\n    - CurseForge\n    - Modrinth\n    - packwiz\n- Detects downloadable resourcepacks and shaders\n- Supports github parsing[¹](#github-rate-limits)\n- Loose modrinth search\n- User friendly toml config\n- Multiple output formats at once\n\n---\n### GitHub rate limits\n\nGitHub has limited requests per hour (up to 60), this means that if you have more than 60 mods, the rest will be excluded from github search. \nIf you authenticate with GitHub using `mmc-export gh-login`, the limit will be removed and requests will be faster. You can always log out with `mmc-export gh-logout`.\n\nIf you don\'t want to use github search by some reason, you can specify `--exclude-providers GitHub` as argument.\n\n# How to Use\n```\nmmc-export -i modpack.zip -c config.toml -f Modrinth packwiz -o converted_modpacks\n```\nIt\'s recommended to fill config at least with basic info like name and version, some launchers can fail import if these values are empty.\n\n## Sub-commands\n\n`gh-login` - to authrize GitHub \\\n`gh-logout` - to get info how to deauthorize GitHub \n\n`purge-cache` - to purge cache. Available arguments:\n```\n--web: to delete requests cache and downloaded mods\n--files: to delete hashes cache\n--all: equivalent to --web --files, deletes all cache (default)\n```\n\n## Syntax\n```\nmmc-export [sub-command] [-h] [-c CONFIG] -i INPUT -f FORMAT [-o OUTPUT] [-v VERSION] [--modrinth-search SEARCH_TYPE] [--exclude-providers PROVIDERS]\n```\n\n### Explanation\n```\n-h --help: prints help\n-i --input: path to modpack, must be zip file exported from MultiMC.\n-c --config: path to config, used to fill the gaps like description or losted mods.\n-f --format: output formats, must be separated by spaces.\n-o --output: directory where converted zip files will be stored.\n-v --version: specify modpack version, will be overriden by config\'s value if exists\n--modrinth-search: modrinth search accuracy\n--exclude-providers: providers you wish to exclude from search\n--skip-cache: don\'t use web cache in this run\n--scheme: output filename formatting scheme, more info in #scheme-formatting\n```\n> All paths can be relative to current working directory or absolute.\n\n`--format` options (case-sensitive): \n- `CurseForge`\n- `Modrinth`\n- `packwiz`\n- `Intermediate` (only for debugging, may contain sensitive data like username)\n\n`--exclude-providers` options (case-sensitive): \n- `CurseForge`\n- `Modrinth`\n- `GitHub`\n\n`--modrinth-search` options:\n- `exact` - by hash (default)\n- `accurate` - by hash or slug\n- `loose` - by hash, slug or long name\n\nThe example for the optional `--config` file [can be found here](example_config.toml). \n\nFor example, if the script says\n\n> No config entry found for resource: ModName\n\nThen you should add **one** of the following entries to the end of the config:\n\n#### Specify source URL\n```\n[[Resource]]\nname = "ModName"\nfilename = "the_name_of_the.jar" \nurl = "https://cdn.modrinth.com/data/abcdefg/versions/1.0.0/the_name_of_the.jar"\n```\n#### Hide the warning\n```\n[[Resource]]\nname = "ModName"\nfilename = "the_name_of_the.jar" \naction = "ignore"\n```\n#### Explicitly move to overrides\n```\n[[Resource]]\nname = "ModName"\nfilename = "the_name_of_the.jar" \naction = "override"\n```\n#### Delete the file altogether\n```\n[[Resource]]\nname = "ModName"\nfilename = "the_name_of_the.jar" \naction = "remove"\n```\n#### To make mod optional you can append\n`optional = true` - to any of above\n\n> Also specifying both name and filename is optional but recommended, you can always leave only one of these\n\n## Scheme Formatting\n\nMust be used as `--scheme "{keyword}_Literally any text"` without file extension, follows python\'s [format string syntax](https://docs.python.org/3/library/string.html#format-string-syntax)\n\n#### Available keywords: \n- `abbr` - provider abbreviation, usually 2 capitals, e.g. `MR`, `CF`\n- `format` - full format name, e.g. `CurseForge`, `Packwiz`\n- `name` same as `pack.name` - modpack name\n- `version` same as `pack.version` - modpack version\n- `pack` - pointer to [Intermediate](mmc_export/Helpers/structures.py#L49-L65) structure\n\nDefault scheme is as simple as `{abbr}_{name}`\n\n***Caution: if you don\'t use any format specifc keywords, output files will overwrite the same file several times***, can be ignored if you output to only one format.\nAlso, be aware of your filesystem limitations, unsupported characters may lead to an error, or inaccesible file.\n\n# How to Install / Update\n```\npip install -U mmc-export\n```\n',
    'author': 'RozeFound',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/RozeFound/mmc-export',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import asyncio
from collections import namedtuple
from datetime import datetime
from json import loads as parse_json
from pathlib import Path
from re import compile as re_compile
from urllib.parse import urlparse
from zipfile import ZipFile

import tenacity as tn
from aiohttp_client_cache.session import CachedSession

from .structures import Intermediate, Resource
from .utils import delete_github_token, get_github_token, get_hash
from .. import config


class ResourceAPI(object):

    modrinth_search_type: str
    excluded_providers: list[str]

    def __init__(self, session: CachedSession, intermediate: Intermediate) -> None:

        self.session = session
        self.intermediate = intermediate

        self.session.headers["User-Agent"] = "RozeFound/mmc-export/2.5.9"
        self.session.headers["X-Api-Key"] = config.CURSEFORGE_API_TOKEN
        self.session.headers["Content-Type"] = "application/json"
        self.session.headers["Accept"] = "application/json"

        self.github = "https://api.github.com"
        self.modrinth = "https://api.modrinth.com/v2"
        self.curseforge = "https://api.curseforge.com/v1"

        self.cache_directory = config.DEFAULT_CACHE_DIR / "v6"
        self.cache_directory.mkdir(parents=True, exist_ok=True)

        super().__init__()

    def _get_raw_info(self, path: Path) -> tuple[dict, Resource]:

        from pickle import HIGHEST_PROTOCOL
        from pickle import dumps as serialize
        from pickle import loads as deserialize

        cache_file = self.cache_directory / get_hash(path, "xxhash")
        if cache_file.exists():
            data = cache_file.read_bytes()
            meta, resource = deserialize(data)
        else:
            meta = {"name": path.stem,
                    "id": None,
                    "version": "0.0.0"}

            if path.suffix in (".jar", ".disabled"):
                with ZipFile(path) as modArchive:
                    filenames = [Path(file).name for file in modArchive.namelist()]
                    if "fabric.mod.json" in filenames:
                        data = modArchive.read("fabric.mod.json")
                        meta = parse_json(data, strict=False)
                    elif "pack.mcmeta" in filenames:
                        data = modArchive.read("pack.mcmeta")
                        json = parse_json(data, strict=False)
                        meta['name'] = json['pack']['description']
                    

            resource = Resource(meta['name'])
            file_data = path.read_bytes()
            resource.file.hash.sha1 = get_hash(file_data, "sha1")
            resource.file.hash.sha256 = get_hash(file_data, "sha256")
            resource.file.hash.sha512 = get_hash(file_data, "sha512")
            resource.file.hash.murmur2 = get_hash(file_data, "murmur2")

            resource.file.size = path.stat().st_size

            to_cache = meta, resource
            data = serialize(to_cache, HIGHEST_PROTOCOL)
            cache_file.write_bytes(data)

        resource.file.path = path
        resource.file.name = path.name
        resource.file.relativePath = path.parent.name

        return meta, resource
    
    @tn.retry(stop=tn.stop.stop_after_attempt(5), wait=tn.wait.wait_fixed(1))
    async def _get_github(self, meta: dict, resource: Resource) -> None:

        if "contact" not in meta or "GitHub" in self.excluded_providers: return
        
        for link in meta['contact'].values():
            parsed_link = urlparse(link)

            if parsed_link.netloc == "github.com":
                owner, repo = parsed_link.path[1:].split('/')[:2]
                repo = repo.removesuffix(".git")
                resource.links.append(f"https://github.com/{owner}/{repo}")
                break
        else: return

        async with self.session.get(f"https://api.github.com/repos/{owner}/{repo}/releases") as response:
            if response.status != 200 and response.status != 504: return

            for release in await response.json():
                for asset in release['assets']:
                    if asset['name'] == resource.file.name:
                        url = asset['browser_download_url']
                        author = release['author']['login']
                        break
                else: continue
                break
            else: return

            resource.providers['Other'] = Resource.Provider(
                ID     = None,
                fileID = None,
                url    = url,
                slug   = meta['id'],
                author = author)


class ResourceAPI_Batched(ResourceAPI):

    def __init__(self, session: CachedSession, intermediate: Intermediate) -> None:

        self.queue: list[tuple[dict, Resource]] = list()

        super().__init__(session, intermediate)

    def queue_resource(self, path: Path) -> None:

        meta, resource = self._get_raw_info(path)
        
        if path.suffix == ".disabled": 
            resource.optional = True
            resource.file.path = path.replace(path.with_suffix(''))
            resource.file.name = resource.file.path.name

        self.queue.append((meta, resource))

    async def gather(self) -> list[Resource]:

        futures = (
            self._get_batched_curseforge(),
            self._get_batched_modrinth(),
            self._get_batched_github()
        )

        await asyncio.gather(*futures)
        resources = [resource for _, resource in self.queue]
        return resources

    @tn.retry(stop=tn.stop.stop_after_attempt(5), wait=tn.wait.wait_fixed(1))
    async def _get_batched_curseforge(self) -> None:

        if "CurseForge" in self.excluded_providers: return

        payload = {"fingerprints":[resource.file.hash.murmur2 for _, resource in self.queue]}
        async with self.session.post(f"{self.curseforge}/fingerprints", json=payload) as response:
            if response.status != 200 and response.status != 504: return
            if matches := (await response.json())['data']['exactMatches']:
                versions = {str(version['file']['fileFingerprint']): version for version in matches}
            else: return

        payload = {"modIds": [version['id'] for version in versions.values()]}
        async with self.session.post(f"{self.curseforge}/mods", json=payload) as response:
            if response.status != 200 and response.status != 504: return
            if addons_array := (await response.json())['data']:
                addons = {addon['id']: addon for addon in addons_array}
            else: return

        for _, resource in self.queue:
            if version := versions.get(resource.file.hash.murmur2):
                if addon := addons.get(version['id']):

                    resource.name = addon['name']
                    resource.links.append(addon['links']['websiteUrl'])
                    if srcUrl := addon['links']['sourceUrl']:
                        resource.links.append(srcUrl)

                    resource.providers['CurseForge'] = Resource.Provider(
                        ID     = addon['id'],
                        fileID = version['file']['id'],
                        url    = version['file']['downloadUrl'],
                        slug   = addon['slug'],
                        author = addon['authors'][0]['name'])

    @tn.retry(stop=tn.stop.stop_after_attempt(5), wait=tn.wait.wait_fixed(1))
    async def _get_batched_modrinth(self) -> None:

        if "Modrinth" in self.excluded_providers: return
        search_queue: list[tuple[dict, Resource]] = list()

        payload = {"algorithm": "sha1", "hashes": [resource.file.hash.sha1 for _, resource in self.queue]}
        async with self.session.post(f"{self.modrinth}/version_files", json=payload) as response:
            if response.status != 200 and response.status != 504 and response.status != 423: return
            versions = await response.json()

            for meta, resource in self.queue:
                if version := versions.get(resource.file.hash.sha1):

                    file = next(file for file in version['files'] 
                        if resource.file.hash.sha1 == file['hashes']['sha1']
                        and resource.file.hash.sha512 == file['hashes']['sha512'])

                    resource.providers['Modrinth'] = Resource.Provider(
                    ID     = version['project_id'],
                    fileID = version['id'],
                    url    = file['url'],
                    slug   = meta['id'])
                else: search_queue.append((meta, resource))

        if self.modrinth_search_type != "exact": await self._get_batched_modrinth_loose(search_queue)

    @tn.retry(stop=tn.stop.stop_after_attempt(5), wait=tn.wait.wait_fixed(1))
    async def _get_batched_modrinth_loose(self, search_queue: list[tuple[dict, Resource]]) -> None:

        version_ids: list[str] = list()
        
        @tn.retry(stop=tn.stop.stop_after_attempt(5), wait=tn.wait.wait_incrementing(1, 15, 60))
        async def get_project_id(meta: dict, resource: Resource):
            if self.modrinth_search_type == "loose":      
                async with self.session.get(f"{self.modrinth}/search?query={resource.name}&limit=1") as response: 
                    if response.status != 200 and response.status != 504 and response.status != 423: return resource, None
                    if hits := (await response.json())['hits']: return resource, hits[0]['project_id']
            return resource, meta['id']

        futures = (get_project_id(meta, resource) for meta, resource in search_queue)
        project_ids = {resource.name: id for resource, id in await asyncio.gather(*futures) if id}
        if not project_ids: return

        l2s = lambda l: "[{}]".format(",".join(map('"{}"'.format, l))) # list to string convesion
        async with self.session.get(f"{self.modrinth}/projects?ids={l2s(project_ids.values())}") as response:
            if response.status != 200 and response.status != 504 and response.status != 423: return
            for project in (projects := await response.json()): version_ids.extend(project['versions'])

            if not version_ids: return

            async with self.session.get(f"{self.modrinth}/versions?ids={l2s(version_ids)}") as response:
                if response.status != 200 and response.status != 504 and response.status != 423: return
                for project in projects: project['versions'] = [version for version in await response.json()
                                                                if version['project_id'] == project['id']]

        minecraft_major_version = ".".join(self.intermediate.minecraft_version.split(".")[:2])
        minecraft_versions = minecraft_major_version, self.intermediate.minecraft_version

        for meta, resource in search_queue:
            if project := next((project for project in projects 
            if project['id'] == project_ids.get(resource.name, "")), None):
                for version in project['versions']:
                    if meta['version'] in version['version_number'] \
                    and self.intermediate.modloader.type in version['loaders'] \
                    and any(mv in version['game_versions'] for mv in minecraft_versions):

                        file = next(file for file in version['files'] 
                            if file['filename'] == resource.file.name or file['primary'])

                        resource.providers['Modrinth'] = Resource.Provider(
                            ID     = version['project_id'],
                            fileID = version['id'],
                            url    = file['url'],
                            slug   = meta['id'])

                        resource.file.hash.sha1 = file['hashes']['sha1']
                        resource.file.hash.sha512 = file['hashes']['sha512']
                        resource.file.size = file['size']

                        break

    @tn.retry(stop=tn.stop.stop_after_attempt(5), wait=tn.wait.wait_fixed(1))
    async def _get_batched_github(self) -> None:

        if "GitHub" in self.excluded_providers: return
        
        if not self.session.headers.get('Authorization'):
            if token := get_github_token(): self.session.headers['Authorization'] = f"Bearer {token}"
            else: 
                futures = [self._get_github(meta, resource) for meta, resource in self.queue]
                await asyncio.gather(*futures)

                async with self.session.disabled():
                    async with self.session.get("https://api.github.com/rate_limit") as response:
                        ratelimit = (await response.json())['resources']['core']
                        time_remaining = datetime.fromtimestamp(float(ratelimit['reset']))
                        if ratelimit['remaining'] == 0: 
                            print("You have exceeded the GitHub API rate-limit, only cached results will be used.")
                            print(f"Please sign in with `mmc-export gh-login` or try again at {time_remaining:%H:%M}")
                return

        Repository = namedtuple('Repository', ['name', 'owner', 'alias'])
        repositories: list[Repository] = list()
        pattern = re_compile(r"[\W_]+")

        for meta, resource in self.queue:
            if "contact" not in meta: continue
            for link in meta['contact'].values():
                parsed_link = urlparse(link)

                if parsed_link.netloc == "github.com":
                    alias = pattern.sub('', meta['id'])
                    owner, name = parsed_link.path[1:].split('/')[:2]
                    repo = Repository(name.removesuffix(".git"), owner, alias)
                    resource.links.append(f"https://github.com/{repo.owner}/{repo.name}")
                    repositories.append(repo)
                    break
            else: continue

        from gql_query_builder import GqlQuery
        queries: list[str] = list()
        
        for repo in repositories:
            query = GqlQuery().fields(['...repoReleaseAssets']) \
                .query('repository', alias=repo.alias, input={"name": f'"{repo.name}"', "owner": f'"{repo.owner}"'}) \
                .generate()
            queries.append(query)

        payload = """
        fragment repoReleaseAssets on Repository {
            releases(last: 100) { edges { node {
                releaseAssets(last: 10) { nodes {
                    name
                    downloadUrl
        } } } } } } """ + GqlQuery().operation(queries=queries).generate()

        async with self.session.post(f"{self.github}/graphql", json={"query": payload}) as response:
            if response.status == 401: delete_github_token(); raise tn.TryAgain
            if response.status != 200 and response.status != 504: return
            data = (await response.json())['data']      

            for meta, resource in self.queue:
                if not data.get(alias := pattern.sub('', meta['id']) if meta['id'] else "unknown"): continue
                for release in data.get(alias, {}).get('releases', {}).get('edges', []):
                    for asset in release.get('node', {}).get('releaseAssets', {}).get('nodes', []):
                        if asset['name'] == resource.file.name: url = asset['downloadUrl']; break
                    else: continue
                    break
                else: continue

                resource.providers['Other'] = Resource.Provider(
                    ID     = None,
                    fileID = None,
                    url    = url,
                    slug   = meta['id'])



from pathlib import Path

from ..Helpers.structures import File, Intermediate, Resource, Writer
from ..Helpers.utils import get_name_from_scheme


class Modrinth(Writer):

    def __init__(self, path: Path, intermediate: Intermediate) -> None:

        self.index = dict()

        super().__init__(path, intermediate)

    def add_resource(self, resource: Resource) -> None:
        
        relative_path = Path(resource.file.relativePath) / resource.file.name

        data = {
            "path": relative_path.as_posix(),
            "hashes": {"sha1": resource.file.hash.sha1, "sha512": resource.file.hash.sha512},
            "downloads": sorted(provider.url for provider in resource.providers.values()),
            "fileSize": resource.file.size
        }

        if resource.optional: data['env'] = {"client": "optional", "server": "optional"}

        self.index['files'].append(data)

    def add_override(self, file: File) -> None:

        overrides_dir = self.temp_dir / "overrides"
        overrides_dir.mkdir(parents=True, exist_ok=True)

        file_path = overrides_dir / file.relativePath
        file_path.mkdir(parents=True, exist_ok=True)

        from shutil import copy2 as copy_file
        copy_file(file.path, file_path)

    def write_index(self) -> None:
        
        match self.intermediate.modloader.type:
            case "fabric": modloader = "fabric-loader"
            case "quilt": modloader = "quilt-loader"
            case _: modloader = self.intermediate.modloader.type
        
        self.index = {
            "formatVersion": 1,
            "game": "minecraft",

            "versionId": self.intermediate.version,
            "name": self.intermediate.name,
            "summary": self.intermediate.description,

            "files": [],

            "dependencies": {
                "minecraft": self.intermediate.minecraft_version,
                modloader: self.intermediate.modloader.version
            }
        }

    def print_bundled(self, bundled_files: list[Resource]) -> None: 

        gh_links = []
        other_links = []

        for resource in bundled_files:

            if resource.links:
                link = sorted(set(resource.links), reverse=True)[-1]
                md_link = f"* [{resource.name}]({link})"
            else: md_link = f"* {resource.name} (link unknown)"; link = ""

            if link.startswith("https://github.com"): 
                gh_links.append(md_link)
            else: other_links.append(md_link)

        newline = '\n'
        message = "Sources for bundled mods:\n\n"
        if gh_links: message +=f"GitHub links\n{newline.join(gh_links)}\n"
        if other_links: message +=f"\nOther links\n{newline.join(other_links)}\n"

        message +="\nAlways check the licenses to see if they allow distribution!"

        md_file = self.modpack_path / "bundled_links.md"
        if gh_links or other_links: md_file.write_text(message)

    def write(self) -> None:

        self.write_index()

        for override in self.intermediate.overrides:
            self.add_override(override)

        bundled_files: list[Resource] = list()
        from copy import deepcopy as copy_object

        for resource in self.intermediate.resources:
            resource_copy = copy_object(resource)
            resource_copy.providers.pop('CurseForge', None)

            if not resource_copy.providers:
                self.add_override(resource.file)
                bundled_files.append(resource)
            else: self.add_resource(resource_copy)

        self.print_bundled(bundled_files)

        from json import dump as write_json
        with open(self.temp_dir / "modrinth.index.json", 'w') as file:
            write_json(self.index, file, indent=4)

        from shutil import make_archive
        name = get_name_from_scheme("MR", "Modrinth", self.intermediate)
        archive = Path(make_archive((self.modpack_path / name).as_posix(), 'zip', self.temp_dir, '.'))
        archive.replace(archive.with_suffix(".mrpack"))
